from django.contrib.sessions.middleware import SessionMiddleware
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings
from django.test.client import RequestFactory
from unittest.mock import MagicMock

from mozilla_django_oidc.utils import (
    absolutify,
    add_state_and_verifier_and_nonce_to_session,
    base64_url_decode,
    base64_url_encode,
    generate_code_challenge,
    import_from_settings,
)


class SettingImportTestCase(TestCase):
    @override_settings(EXAMPLE_VARIABLE="example_value")
    def test_attr_existing_no_default_value(self):
        s = import_from_settings("EXAMPLE_VARIABLE")
        self.assertEqual(s, "example_value")

    def test_attr_nonexisting_no_default_value(self):
        with self.assertRaises(ImproperlyConfigured):
            import_from_settings("EXAMPLE_VARIABLE")

    def test_attr_nonexisting_default_value(self):
        s = import_from_settings("EXAMPLE_VARIABLE", "example_default")
        self.assertEqual(s, "example_default")


class AbsolutifyTestCase(TestCase):
    def test_absolutify(self):
        req = RequestFactory().get("/something/else")
        url = absolutify(req, "/foo/bar")
        self.assertEqual(url, "http://testserver/foo/bar")

        req = RequestFactory().get("/something/else", SERVER_PORT=8888)
        url = absolutify(req, "/foo/bar")
        self.assertEqual(url, "http://testserver:8888/foo/bar")

    @override_settings(SECURE_PROXY_SSL_HEADER=("HTTP_X_FORWARDED_PROTO", "https"))
    def test_absolutify_https(self):
        req = RequestFactory(HTTP_X_FORWARDED_PROTO="https").get("/", SERVER_PORT=443)
        url = absolutify(req, "/foo/bar")
        self.assertEqual(url, "https://testserver/foo/bar")

    @override_settings(SECURE_PROXY_SSL_HEADER=("HTTP_X_FORWARDED_PROTO", "https"))
    def test_absolutify_path_host_injection(self):
        req = RequestFactory(HTTP_X_FORWARDED_PROTO="https").get("/", SERVER_PORT=443)
        url = absolutify(req, "evil.com/foo/bar")
        self.assertEqual(url, "https://testserver/evil.com/foo/bar")


class Base64URLEncodeTestCase(TestCase):
    def test_base64_url_encode(self):
        """
        Tests creating a url-safe base64 encoded string from bytes.
        Source: https://datatracker.ietf.org/doc/html/rfc7636#appendix-A
        """
        data = bytes((3, 236, 255, 224, 193))
        encoded = base64_url_encode(data)

        # Using base64.b64encode() returns b'A+z/4ME='.
        # Our implementation should strip tailing '='s padding.
        # and replace '+' with '-' and '/' with '_'.
        self.assertEqual(encoded, "A-z_4ME")

        # Decoding should return the original data.
        decoded = base64_url_decode(encoded)
        self.assertEqual(decoded, data)

    def test_base64_url_encode_empty_input(self):
        """
        Tests creating a url-safe base64 encoded string from an empty bytes instance.
        """
        data = bytes()
        encoded = base64_url_encode(data)
        self.assertEqual(encoded, "")

        decoded = base64_url_decode(encoded)
        self.assertEqual(decoded, data)

    def test_base64_url_encode_double_padding(self):
        """
        Test encoding a string whoose base64.b64encode encoding ends with '=='.
        """
        data = bytes((3, 236, 255, 224, 193, 222, 22))
        encoded = base64_url_encode(data)

        # Using base64.b64encode() returns b'A+z/4MHeFg=='.
        self.assertEqual(encoded, "A-z_4MHeFg")

        # Decoding should return the original data.
        decoded = base64_url_decode(encoded)
        self.assertEqual(decoded, data)


class PKCECodeVerificationTestCase(TestCase):
    def test_generate_code_challenge(self):
        """
        Tests that a code challenge is generated correctly with the 'S256' method.
        """
        code_verifier = "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"
        code_challenge = generate_code_challenge(code_verifier, "S256")

        self.assertEqual(code_challenge, "E9Melhoa2OwvFrEMTJguCHaoeK1t8URWbuGJSstw-cM")

    def test_generate_plain_code_challenge(self):
        """
        Tests that a code challenge is generated correctly with the 'plain' method.
        """
        code_verifier = "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"
        code_challenge = generate_code_challenge(code_verifier, "plain")

        self.assertEqual(code_challenge, code_verifier)

    def test_generate_code_challenge_invalid_method(self):
        """
        Tests that an exception is raised when an invalid code challenge method is provided.
        """
        code_verifier = "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"
        self.assertRaises(ValueError, generate_code_challenge, code_verifier, "INVALID")


class SessionStateTestCase(TestCase):
    def setUp(self):
        self.request = RequestFactory().get("/doesnt/matter")

        # Setup request with a session for testing
        middleware = SessionMiddleware(MagicMock())
        middleware.process_request(self.request)
        self.request.session.save()

    def test_add_state_to_session(self):
        state = "example_state"
        params = {}

        add_state_and_verifier_and_nonce_to_session(self.request, state, params)

        self.assertIn("oidc_states", self.request.session)
        self.assertEqual(1, len(self.request.session["oidc_states"]))
        self.assertIn(state, self.request.session["oidc_states"].keys())

    def test_multiple_states(self):
        state1 = "example_state_1"
        state2 = "example_state_2"
        params = {}

        add_state_and_verifier_and_nonce_to_session(self.request, state1, params)

        self.assertEqual(1, len(self.request.session["oidc_states"]))
        self.assertIn(state1, self.request.session["oidc_states"].keys())

        add_state_and_verifier_and_nonce_to_session(self.request, state2, params)

        self.assertEqual(2, len(self.request.session["oidc_states"]))
        self.assertIn(state1, self.request.session["oidc_states"].keys())
        self.assertIn(state2, self.request.session["oidc_states"].keys())

    def test_max_states(self):
        limit = import_from_settings("OIDC_MAX_STATES", 50)

        first_state = "example_state_0"
        params = {}
        for i in range(limit):
            state = "example_state_{}".format(i)
            add_state_and_verifier_and_nonce_to_session(self.request, state, params)

        self.assertEqual(limit, len(self.request.session["oidc_states"]))
        self.assertIn(first_state, self.request.session["oidc_states"])

        # Add another state which should remove the very first one
        additional_state = "example_state"
        add_state_and_verifier_and_nonce_to_session(
            self.request, additional_state, params
        )

        # Make sure the oldest state was deleted
        self.assertNotIn(first_state, self.request.session["oidc_states"])

        # New state should be in the list but length should not have changed
        self.assertEqual(limit, len(self.request.session["oidc_states"]))
        self.assertIn(additional_state, self.request.session["oidc_states"].keys())

    @override_settings(OIDC_USE_NONCE=False)
    def test_state_dictionary_without_nonce_format(self):
        state = "example_state"
        params = {}

        add_state_and_verifier_and_nonce_to_session(self.request, state, params)

        # Test state dictionary
        self.assertIn(state, self.request.session["oidc_states"].keys())
        self.assertTrue(isinstance(self.request.session["oidc_states"][state], dict))

        # Test nonce
        self.assertIsNone(self.request.session["oidc_states"][state]["nonce"])

        # Test added_on timestamp
        self.assertTrue(
            isinstance(self.request.session["oidc_states"][state]["added_on"], float)
        )

    def test_state_dictionary_with_nonce_format(self):
        state = "example_state"
        params = {"nonce": "example_nonce"}

        add_state_and_verifier_and_nonce_to_session(self.request, state, params)

        self.assertNotEqual(self.request.session["oidc_states"][state]["nonce"], None)



import json
from unittest.mock import Mock, call, patch

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac, serialization
from cryptography.hazmat.primitives.asymmetric import ec
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured, SuspiciousOperation
from django.test import RequestFactory, TestCase, override_settings
from django.utils.encoding import force_bytes, smart_str
from josepy.b64 import b64encode
from josepy.jwa import ES256

from mozilla_django_oidc.auth import OIDCAuthenticationBackend, default_username_algo

User = get_user_model()


class DefaultUsernameAlgoTestCase(TestCase):
    def run_test(self, data, expected):
        actual = default_username_algo(data)
        self.assertEqual(actual, expected)
        self.assertEqual(type(actual), type(expected))

    def test_empty(self):
        self.run_test("", "2jmj7l5rSw0yVb_vlWAYkK_YBwk")

    def test_email(self):
        self.run_test("janet@example.com", "VUCUpl08JVpFeAFKBYkAjLhsQ1c")


class OIDCAuthenticationBackendTestCase(TestCase):
    """Authentication tests."""

    @override_settings(OIDC_OP_TOKEN_ENDPOINT="https://server.example.com/token")
    @override_settings(OIDC_OP_USER_ENDPOINT="https://server.example.com/user")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_RP_CLIENT_SECRET="client_secret")
    def setUp(self):
        self.backend = OIDCAuthenticationBackend()

    def test_missing_request_arg(self):
        """Test authentication returns `None` when `request` is not provided."""
        self.assertIsNone(self.backend.authenticate(request=None))

    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc.auth.requests")
    def test_invalid_token(self, request_mock, token_mock):
        """Test authentication with an invalid token."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        token_mock.return_value = None
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "accesss_token": "access_token",
        }
        request_mock.post.return_value = post_json_mock
        self.assertIsNone(self.backend.authenticate(request=auth_request))

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=True)
    def test_allowed_unsecured_token(self):
        """Test payload data from unsecured token (allowed)."""
        header = force_bytes(json.dumps({"alg": "none"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))
        signature = ""
        token = force_bytes(
            "{}.{}.{}".format(
                smart_str(b64encode(header)), smart_str(b64encode(payload)), signature
            )
        )

        extracted_payload = self.backend.get_payload_data(token, None)
        self.assertEqual(payload, extracted_payload)

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=False)
    def test_disallowed_unsecured_token(self):
        """Test payload data from unsecured token (disallowed)."""
        header = force_bytes(json.dumps({"alg": "none"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))
        signature = ""
        token = force_bytes(
            "{}.{}.{}".format(
                smart_str(b64encode(header)), smart_str(b64encode(payload)), signature
            )
        )

        with self.assertRaises(KeyError):
            self.backend.get_payload_data(token, None)

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=True)
    def test_allowed_unsecured_valid_token(self):
        """Test payload data from valid secured token (unsecured allowed)."""
        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )
        token_bytes = force_bytes(token)
        key_text = smart_str(key)
        output = self.backend.get_payload_data(token_bytes, key_text)
        self.assertEqual(output, payload)

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=False)
    def test_disallowed_unsecured_valid_token(self):
        """Test payload data from valid secure token (unsecured disallowed)."""
        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )
        token_bytes = force_bytes(token)
        key_text = smart_str(key)
        output = self.backend.get_payload_data(token_bytes, key_text)
        self.assertEqual(output, payload)

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=True)
    def test_allowed_unsecured_invalid_token(self):
        """Test payload data from invalid secure token (unsecured allowed)."""
        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        fake_key = b"mysupersecurefaketestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )
        token_bytes = force_bytes(token)
        key_text = smart_str(fake_key)

        with self.assertRaises(SuspiciousOperation) as ctx:
            self.backend.get_payload_data(token_bytes, key_text)
        self.assertEqual(ctx.exception.args[0], "JWS token verification failed.")

    @override_settings(OIDC_ALLOW_UNSECURED_JWT=False)
    def test_disallowed_unsecured_invalid_token(self):
        """Test payload data from invalid secure token (unsecured disallowed)."""
        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        fake_key = b"mysupersecurefaketestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )
        token_bytes = force_bytes(token)
        key_text = smart_str(fake_key)

        with self.assertRaises(SuspiciousOperation) as ctx:
            self.backend.get_payload_data(token_bytes, key_text)
        self.assertEqual(ctx.exception.args[0], "JWS token verification failed.")

    def test_get_user(self):
        """Test get_user method with valid user."""

        user = User.objects.create_user("example_username")
        self.assertEqual(self.backend.get_user(user.pk), user)

    def test_get_invalid_user(self):
        """Test get_user method with non existing user."""

        self.assertIsNone(self.backend.get_user(user_id=1))

    @override_settings(ROOT_URLCONF="tests.namespaced_urls")
    @override_settings(
        OIDC_AUTHENTICATION_CALLBACK_URL="namespace:oidc_authentication_callback"
    )
    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    def test_successful_authentication_existing_user_namespaced(
        self, token_mock, request_mock
    ):
        """Test successful authentication for existing user."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        user = User.objects.create_user(
            username="a_username", email="email@example.com"
        )
        token_mock.return_value = True
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock

        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/namespace/callback/",
        }
        self.assertEqual(self.backend.authenticate(request=auth_request), user)
        token_mock.assert_called_once_with("id_token", nonce=None)
        request_mock.post.assert_called_once_with(
            "https://server.example.com/token",
            data=post_data,
            auth=None,
            verify=True,
            timeout=None,
            proxies=None,
        )
        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )

    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    def test_successful_authentication_existing_user(self, token_mock, request_mock):
        """Test successful authentication for existing user."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        user = User.objects.create_user(
            username="a_username", email="email@example.com"
        )
        token_mock.return_value = True
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock

        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/callback/",
        }
        self.assertEqual(self.backend.authenticate(request=auth_request), user)
        token_mock.assert_called_once_with("id_token", nonce=None)
        request_mock.post.assert_called_once_with(
            "https://server.example.com/token",
            data=post_data,
            auth=None,
            verify=True,
            timeout=None,
            proxies=None,
        )
        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )

    @override_settings(OIDC_STORE_ACCESS_TOKEN=True)
    @override_settings(OIDC_STORE_ID_TOKEN=True)
    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    def test_successful_authentication_existing_user_upper_case(
        self, token_mock, request_mock
    ):
        """Test successful authentication for existing user regardless of email case."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        user = User.objects.create_user(
            username="a_username", email="EMAIL@EXAMPLE.COM"
        )
        token_mock.return_value = True
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock

        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/callback/",
        }
        self.assertEqual(self.backend.authenticate(request=auth_request), user)
        token_mock.assert_called_once_with("id_token", nonce=None)
        request_mock.post.assert_called_once_with(
            "https://server.example.com/token",
            data=post_data,
            auth=None,
            verify=True,
            timeout=None,
            proxies=None,
        )
        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )
        self.assertEqual(auth_request.session.get("oidc_id_token"), "id_token")
        self.assertEqual(
            auth_request.session.get("oidc_access_token"), "access_granted"
        )

    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_claims")
    def test_failed_authentication_verify_claims(
        self, claims_mock, token_mock, request_mock
    ):
        """Test successful authentication for existing user."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        User.objects.create_user(username="a_username", email="email@example.com")
        token_mock.return_value = True
        claims_mock.return_value = False
        get_json_mock = Mock()
        claims_response = {"nickname": "a_username", "email": "email@example.com"}
        get_json_mock.json.return_value = claims_response
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock

        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/callback/",
        }
        self.assertIsNone(self.backend.authenticate(request=auth_request))
        token_mock.assert_called_once_with("id_token", nonce=None)
        claims_mock.assert_called_once_with(claims_response)
        request_mock.post.assert_called_once_with(
            "https://server.example.com/token",
            data=post_data,
            auth=None,
            verify=True,
            timeout=None,
            proxies=None,
        )
        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )

    @patch.object(settings, "OIDC_USERNAME_ALGO")
    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    def test_successful_authentication_new_user(
        self, token_mock, request_mock, algo_mock
    ):
        """Test successful authentication and user creation."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        algo_mock.return_value = "username_algo"
        token_mock.return_value = True
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/callback/",
        }
        self.assertEqual(User.objects.all().count(), 0)
        self.backend.authenticate(request=auth_request)
        self.assertEqual(User.objects.all().count(), 1)
        user = User.objects.all()[0]
        self.assertEqual(user.email, "email@example.com")
        self.assertEqual(user.username, "username_algo")

        token_mock.assert_called_once_with("id_token", nonce=None)
        request_mock.post.assert_called_once_with(
            "https://server.example.com/token",
            data=post_data,
            auth=None,
            verify=True,
            timeout=None,
            proxies=None,
        )
        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )

    @override_settings(OIDC_TOKEN_USE_BASIC_AUTH=True)
    @override_settings(OIDC_STORE_ACCESS_TOKEN=True)
    @override_settings(OIDC_STORE_ID_TOKEN=True)
    @patch("mozilla_django_oidc.auth.requests")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.verify_token")
    def test_successful_authentication_basic_auth_token(self, token_mock, request_mock):
        """
        Test successful authentication when using HTTP basic authentication
        for token endpoint authentication.
        """
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        user = User.objects.create_user(
            username="a_username", email="EMAIL@EXAMPLE.COM"
        )
        token_mock.return_value = True
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock

        post_data = {
            "client_id": "example_id",
            "client_secret": "client_secret",
            "grant_type": "authorization_code",
            "code": "foo",
            "redirect_uri": "http://testserver/callback/",
        }
        self.assertEqual(self.backend.authenticate(request=auth_request), user)
        token_mock.assert_called_once_with("id_token", nonce=None)

        # As the auth parameter is and object, we can't compare them directly
        request_mock.post.assert_called_once()
        post_params = request_mock.post.call_args
        _kwargs = post_params[1]

        self.assertEqual(post_params[0][0], "https://server.example.com/token")
        # Test individual params separately
        sent_data = _kwargs["data"]
        self.assertEqual(sent_data["client_id"], post_data["client_id"])
        self.assertTrue("client_secret" not in _kwargs["data"])
        self.assertEqual(sent_data["grant_type"], post_data["grant_type"])
        self.assertEqual(sent_data["code"], post_data["code"])
        self.assertEqual(sent_data["redirect_uri"], post_data["redirect_uri"])

        auth = _kwargs["auth"]  # requests.auth.HTTPBasicAuth
        self.assertEqual(auth.username, "example_id")
        self.assertEqual(auth.password, "client_secret")
        self.assertEqual(_kwargs["verify"], True)

        request_mock.get.assert_called_once_with(
            "https://server.example.com/user",
            headers={"Authorization": "Bearer access_granted"},
            verify=True,
            timeout=None,
            proxies=None,
        )
        self.assertEqual(auth_request.session.get("oidc_id_token"), "id_token")
        self.assertEqual(
            auth_request.session.get("oidc_access_token"), "access_granted"
        )

    def test_authenticate_no_code_no_state(self):
        """Test authenticate with wrong parameters."""

        # there are no GET params
        request = RequestFactory().get("/foo")
        request.session = {}
        self.assertIsNone(self.backend.authenticate(request=request))

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_jwt_decode_params(self, request_mock, jws_mock):
        """Test jwt verification signature."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        jws_mock.return_value = json.dumps({"aud": "audience"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "token",
            "access_token": "access_token",
        }
        request_mock.post.return_value = post_json_mock
        self.backend.authenticate(request=auth_request)
        calls = [call(force_bytes("token"), "client_secret")]
        jws_mock.assert_has_calls(calls)

    @override_settings(OIDC_VERIFY_JWT=False)
    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_jwt_decode_params_verify_false(self, request_mock, jws_mock):
        """Test jwt verification signature with verify False"""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        jws_mock.return_value = json.dumps({"aud": "audience"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "token",
            "access_token": "access_token",
        }
        request_mock.post.return_value = post_json_mock
        calls = [call(force_bytes("token"), "client_secret")]
        self.backend.authenticate(request=auth_request)
        jws_mock.assert_has_calls(calls)

    @override_settings(OIDC_USE_NONCE=True)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    def test_jwt_failed_nonce(self, jws_mock):
        """Test Nonce verification."""

        jws_mock.return_value = json.dumps({"nonce": "foobar", "aud": "aud"}).encode(
            "utf-8"
        )
        id_token = "my_token"
        with self.assertRaisesMessage(
            SuspiciousOperation, "JWT Nonce verification failed."
        ):
            self.backend.verify_token(id_token, **{"nonce": "foo"})

    @override_settings(OIDC_CREATE_USER=False)
    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_create_user_disabled(self, request_mock, jws_mock):
        """Test with user creation disabled and no user found."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertIsNone(self.backend.authenticate(request=auth_request))

    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    @override_settings(OIDC_USE_NONCE=False)
    def test_create_user_enabled(self, request_mock, jws_mock):
        """Test with user creation enabled and no user found."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        self.assertEqual(User.objects.filter(email="email@example.com").exists(), False)
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertEqual(
            self.backend.authenticate(request=auth_request),
            User.objects.get(email="email@example.com"),
        )

    @patch.object(settings, "OIDC_USERNAME_ALGO")
    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_custom_username_algo(self, request_mock, jws_mock, algo_mock):
        """Test user creation with custom username algorithm."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        self.assertEqual(User.objects.filter(email="email@example.com").exists(), False)
        algo_mock.return_value = "username_algo"
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertEqual(
            self.backend.authenticate(request=auth_request),
            User.objects.get(username="username_algo"),
        )

    @override_settings(
        OIDC_USE_NONCE=False,
        OIDC_USERNAME_ALGO="tests.test_auth.dotted_username_algo_callback",
    )
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_custom_username_algo_dotted_path(self, request_mock, jws_mock):
        """Test user creation with custom username algorithm with a dotted path."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        self.assertEqual(User.objects.filter(email="email@example.com").exists(), False)
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertEqual(
            self.backend.authenticate(request=auth_request),
            User.objects.get(username="dotted_username_algo"),
        )

    @override_settings(
        OIDC_USE_NONCE=False,
        OIDC_USERNAME_ALGO="tests.test_auth.dotted_username_algo_callback_with_claims",
    )
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_dotted_username_algo_callback_with_claims(self, request_mock, jws_mock):
        """Test user creation with custom username algorithm with a dotted path."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        self.assertEqual(User.objects.filter(email="email@example.com").exists(), False)
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        domain = "django.con"
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
            "domain": domain,
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertEqual(
            self.backend.authenticate(request=auth_request),
            User.objects.get(username=f"{domain}/email@example.com"),
        )

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_duplicate_emails_exact(self, request_mock, jws_mock):
        """Test auth with two users having the same email."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        User.objects.create(username="user1", email="email@example.com")
        User.objects.create(username="user2", email="email@example.com")
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertIsNone(self.backend.authenticate(request=auth_request))

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_duplicate_emails_case_mismatch(self, request_mock, jws_mock):
        """Test auth with two users having the same email, with different case."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        User.objects.create(username="user1", email="email@example.com")
        User.objects.create(username="user2", email="eMaIl@ExAmPlE.cOm")
        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertIsNone(self.backend.authenticate(request=auth_request))

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.update_user")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_custom_update_user(self, request_mock, jws_mock, update_user_mock):
        """User updated with new claims"""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        User.objects.create(
            username="user1", email="email@example.com", first_name="User"
        )

        def update_user(user, claims):
            user.first_name = claims["nickname"]
            user.save()

        update_user_mock.side_effect = update_user

        jws_mock.return_value = json.dumps({"nonce": "nonce"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "a_username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "id_token",
            "access_token": "access_granted",
        }
        request_mock.post.return_value = post_json_mock
        self.assertIsNone(self.backend.authenticate(request=auth_request))

        self.assertEqual(User.objects.get().first_name, "a_username")


class OIDCAuthenticationBackendRS256WithKeyTestCase(TestCase):
    """Authentication tests with ALG RS256 and provided IdP Sign Key."""

    @override_settings(OIDC_OP_TOKEN_ENDPOINT="https://server.example.com/token")
    @override_settings(OIDC_OP_USER_ENDPOINT="https://server.example.com/user")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_RP_CLIENT_SECRET="client_secret")
    @override_settings(OIDC_RP_SIGN_ALGO="RS256")
    @override_settings(OIDC_RP_IDP_SIGN_KEY="sign_key")
    def setUp(self):
        self.backend = OIDCAuthenticationBackend()

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.requests")
    def test_jwt_verify_sign_key(self, request_mock, jws_mock):
        """Test jwt verification signature."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        jws_mock.return_value = json.dumps({"aud": "audience"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "token",
            "access_token": "access_token",
        }
        request_mock.post.return_value = post_json_mock
        self.backend.authenticate(request=auth_request)
        calls = [call(force_bytes("token"), "sign_key")]
        jws_mock.assert_has_calls(calls)


class OIDCAuthenticationBackendRS256WithJwksEndpointTestCase(TestCase):
    """Authentication tests with ALG RS256 and IpD JWKS Endpoint."""

    @override_settings(OIDC_OP_TOKEN_ENDPOINT="https://server.example.com/token")
    @override_settings(OIDC_OP_USER_ENDPOINT="https://server.example.com/user")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_RP_CLIENT_SECRET="client_secret")
    @override_settings(OIDC_RP_SIGN_ALGO="RS256")
    @override_settings(OIDC_OP_JWKS_ENDPOINT="https://server.example.com/jwks")
    def setUp(self):
        self.backend = OIDCAuthenticationBackend()

    @override_settings(OIDC_USE_NONCE=False)
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend._verify_jws")
    @patch("mozilla_django_oidc.auth.OIDCAuthenticationBackend.retrieve_matching_jwk")
    @patch("mozilla_django_oidc.auth.requests")
    def test_jwt_verify_sign_key_calls(self, request_mock, jwk_mock, jws_mock):
        """Test jwt verification signature."""
        auth_request = RequestFactory().get("/foo", {"code": "foo", "state": "bar"})
        auth_request.session = {}

        jwk_mock_ret = {
            "kty": "RSA",
            "alg": "RS256",
            "use": "sig",
            "kid": "cc7d29c9cb3780741cc0876633c9107a0f33c289",
            "n": "20LvblCBaPicNV3-NnJuahqbpi-b8hFD",
            "e": "AQAB",
        }
        jwk_mock.return_value = jwk_mock_ret

        jws_mock.return_value = json.dumps({"aud": "audience"}).encode("utf-8")
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "nickname": "username",
            "email": "email@example.com",
        }
        request_mock.get.return_value = get_json_mock
        post_json_mock = Mock(status_code=200)
        post_json_mock.json.return_value = {
            "id_token": "token",
            "access_token": "access_token",
        }
        request_mock.post.return_value = post_json_mock
        self.backend.authenticate(request=auth_request)
        calls = [call(force_bytes("token"), jwk_mock_ret)]
        jws_mock.assert_has_calls(calls)

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_matching_jwk(self, mock_requests):
        """Test retrieving valid jwk"""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "alg": "RS256",
                    "kid": "foobar",
                },
                {
                    "alg": "RS512",
                    "kid": "foobar512",
                },
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(
            json.dumps({"alg": "RS256", "typ": "JWT", "kid": "foobar"})
        )
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        jwk_key = self.backend.retrieve_matching_jwk(force_bytes(token))
        self.assertEqual(jwk_key, get_json_mock.json.return_value["keys"][0])

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_matching_jwk_same_kid(self, mock_requests):
        """Test retrieving valid jwk from a list of keys with the same kid"""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "alg": "RS512",
                    "kid": "foobar",
                },
                {
                    "alg": "RS384",
                    "kid": "foobar",
                },
                {
                    "alg": "RS256",
                    "kid": "foobar",
                },
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(
            json.dumps({"alg": "RS256", "typ": "JWT", "kid": "foobar"})
        )
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        jwk_key = self.backend.retrieve_matching_jwk(force_bytes(token))
        self.assertEqual(jwk_key, get_json_mock.json.return_value["keys"][2])

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_mismatcing_jwk_alg(self, mock_requests):
        """Test retrieving mismatching jwk alg"""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "alg": "foo",
                    "kid": "bar",
                }
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT", "kid": "bar"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        with self.assertRaises(SuspiciousOperation) as ctx:
            self.backend.retrieve_matching_jwk(force_bytes(token))

        self.assertEqual(ctx.exception.args[0], "Could not find a valid JWKS.")

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_mismatcing_jwk_kid(self, mock_requests):
        """Test retrieving mismatching jwk kid"""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "alg": "HS256",
                    "kid": "foobar",
                }
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT", "kid": "bar"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        with self.assertRaises(SuspiciousOperation) as ctx:
            self.backend.retrieve_matching_jwk(force_bytes(token))

        self.assertEqual(ctx.exception.args[0], "Could not find a valid JWKS.")

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_jwk_optional_alg(self, mock_requests):
        """Test retrieving jwk with optional alg"""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "kid": "kid",
                }
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(json.dumps({"alg": "HS256", "typ": "JWT", "kid": "kid"}))
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        jwk_key = self.backend.retrieve_matching_jwk(force_bytes(token))
        self.assertEqual(jwk_key, get_json_mock.json.return_value["keys"][0])

    @patch("mozilla_django_oidc.auth.requests")
    def test_retrieve_not_existing_jwk(self, mock_requests):
        """Test retrieving jwk that doesn't exist."""

        get_json_mock = Mock()
        get_json_mock.json.return_value = {"keys": [{"alg": "RS256", "kid": "kid"}]}
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(
            json.dumps({"alg": "RS256", "typ": "JWT", "kid": "differentkid"})
        )
        payload = force_bytes(json.dumps({"foo": "bar"}))

        # Compute signature
        key = b"mysupersecuretestkey"
        h = hmac.HMAC(key, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)), smart_str(b64encode(payload))
        )
        h.update(force_bytes(msg))
        signature = b64encode(h.finalize())

        token = "{}.{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(payload)),
            smart_str(signature),
        )

        with self.assertRaises(SuspiciousOperation) as ctx:
            self.backend.retrieve_matching_jwk(force_bytes(token))

        self.assertEqual(ctx.exception.args[0], "Could not find a valid JWKS.")


class TestVerifyClaim(TestCase):
    @patch("mozilla_django_oidc.auth.import_from_settings")
    def test_returns_false_if_email_not_in_claims(self, patch_settings):
        patch_settings.return_value = "openid email"
        ret = OIDCAuthenticationBackend().verify_claims({})
        self.assertFalse(ret)

    @patch("mozilla_django_oidc.auth.import_from_settings")
    def test_returns_true_if_email_in_claims(self, patch_settings):
        patch_settings.return_value = "openid email"
        ret = OIDCAuthenticationBackend().verify_claims({"email": "email@example.com"})
        self.assertTrue(ret)

    @patch("mozilla_django_oidc.auth.import_from_settings")
    @patch("mozilla_django_oidc.auth.LOGGER")
    def test_returns_true_custom_claims(self, patch_logger, patch_settings):
        patch_settings.return_value = "foo bar"
        ret = OIDCAuthenticationBackend().verify_claims({})
        self.assertTrue(ret)
        msg = (
            "Custom OIDC_RP_SCOPES defined. "
            "You need to override `verify_claims` for custom claims verification."
        )
        patch_logger.warning.assert_called_with(msg)


def dotted_username_algo_callback(email):
    return "dotted_username_algo"


def dotted_username_algo_callback_with_claims(email, claims=None):
    domain = claims["domain"]
    username = f"{domain}/{email}"
    return username


@override_settings(OIDC_OP_TOKEN_ENDPOINT="https://server.example.com/token")
@override_settings(OIDC_OP_USER_ENDPOINT="https://server.example.com/user")
@override_settings(OIDC_RP_CLIENT_ID="example_id")
@override_settings(OIDC_RP_CLIENT_SECRET="client_secret")
@override_settings(OIDC_RP_SIGN_ALGO="ES256")
class OIDCAuthenticationBackendES256WithJwksEndpointTestCase(TestCase):
    """Authentication tests with ALG ES256 and IpD JWKS Endpoint."""

    def test_es256_alg_misconfiguration(self):
        """Test that ES algorithm requires a JWKS endpoint"""

        with self.assertRaises(ImproperlyConfigured) as ctx:
            OIDCAuthenticationBackend()

        self.assertEqual(
            ctx.exception.args[0],
            "ES256 alg requires OIDC_RP_IDP_SIGN_KEY or OIDC_OP_JWKS_ENDPOINT to be configured.",
        )

    @patch("mozilla_django_oidc.auth.requests")
    @override_settings(OIDC_OP_JWKS_ENDPOINT="https://server.example.com/jwks")
    def test_es256_alg_verification(self, mock_requests):
        """Test that token can be verified with the ES algorithm"""

        self.backend = OIDCAuthenticationBackend()

        # Generate a private key to create a test token with
        private_key = ec.generate_private_key(ec.SECP256R1, default_backend())
        private_key_pem = private_key.private_bytes(
            serialization.Encoding.PEM,
            serialization.PrivateFormat.PKCS8,
            serialization.NoEncryption(),
        )

        # Make the public key available through the JWKS response
        public_numbers = private_key.public_key().public_numbers()
        get_json_mock = Mock()
        get_json_mock.json.return_value = {
            "keys": [
                {
                    "kid": "eckid",
                    "kty": "EC",
                    "alg": "ES256",
                    "use": "sig",
                    "x": smart_str(b64encode(public_numbers.x.to_bytes(32, "big"))),
                    "y": smart_str(b64encode(public_numbers.y.to_bytes(32, "big"))),
                    "crv": "P-256",
                }
            ]
        }
        mock_requests.get.return_value = get_json_mock

        header = force_bytes(
            json.dumps(
                {
                    "typ": "JWT",
                    "alg": "ES256",
                    "kid": "eckid",
                },
            )
        )
        data = {"name": "John Doe", "test": "test_es256_alg_verification"}

        h = hmac.HMAC(private_key_pem, hashes.SHA256(), backend=default_backend())
        msg = "{}.{}".format(
            smart_str(b64encode(header)),
            smart_str(b64encode(force_bytes(json.dumps(data)))),
        )
        h.update(force_bytes(msg))

        signature = b64encode(ES256.sign(private_key, force_bytes(msg)))
        token = "{}.{}".format(
            msg,
            smart_str(signature),
        )

        # Verify the token created with the private key by using the JWKS endpoint,
        # where the public numbers are.
        payload = self.backend.verify_token(token)

        self.assertEqual(payload, data)
        mock_requests.get.assert_called_once()



from django.urls import path
from django.utils.module_loading import import_string

from mozilla_django_oidc import views
from mozilla_django_oidc.utils import import_from_settings

DEFAULT_CALLBACK_CLASS = "mozilla_django_oidc.views.OIDCAuthenticationCallbackView"
CALLBACK_CLASS_PATH = import_from_settings(
    "OIDC_CALLBACK_CLASS", DEFAULT_CALLBACK_CLASS
)

OIDCCallbackClass = import_string(CALLBACK_CLASS_PATH)


DEFAULT_AUTHENTICATE_CLASS = "mozilla_django_oidc.views.OIDCAuthenticationRequestView"
AUTHENTICATE_CLASS_PATH = import_from_settings(
    "OIDC_AUTHENTICATE_CLASS", DEFAULT_AUTHENTICATE_CLASS
)

OIDCAuthenticateClass = import_string(AUTHENTICATE_CLASS_PATH)

urlpatterns = [
    path("callback/", OIDCCallbackClass.as_view(), name="oidc_authentication_callback"),
    path(
        "authenticate/",
        OIDCAuthenticateClass.as_view(),
        name="oidc_authentication_init",
    ),
    path("logout/", views.OIDCLogoutView.as_view(), name="oidc_logout"),
]



import logging
import time
from re import Pattern as re_Pattern
from urllib.parse import quote, urlencode

from django.contrib.auth import BACKEND_SESSION_KEY
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import cached_property
from django.utils.module_loading import import_string

from mozilla_django_oidc.auth import OIDCAuthenticationBackend
from mozilla_django_oidc.utils import (
    absolutify,
    add_state_and_verifier_and_nonce_to_session,
    generate_code_challenge,
    import_from_settings,
)

LOGGER = logging.getLogger(__name__)


class SessionRefresh(MiddlewareMixin):
    """Refreshes the session with the OIDC RP after expiry seconds

    For users authenticated with the OIDC RP, verify tokens are still valid and
    if not, force the user to re-authenticate silently.

    """

    def __init__(self, get_response):
        super(SessionRefresh, self).__init__(get_response)
        self.OIDC_EXEMPT_URLS = self.get_settings("OIDC_EXEMPT_URLS", [])
        self.OIDC_OP_AUTHORIZATION_ENDPOINT = self.get_settings(
            "OIDC_OP_AUTHORIZATION_ENDPOINT"
        )
        self.OIDC_RP_CLIENT_ID = self.get_settings("OIDC_RP_CLIENT_ID")
        self.OIDC_STATE_SIZE = self.get_settings("OIDC_STATE_SIZE", 32)
        self.OIDC_AUTHENTICATION_CALLBACK_URL = self.get_settings(
            "OIDC_AUTHENTICATION_CALLBACK_URL",
            "oidc_authentication_callback",
        )
        self.OIDC_RP_SCOPES = self.get_settings("OIDC_RP_SCOPES", "openid email")
        self.OIDC_USE_NONCE = self.get_settings("OIDC_USE_NONCE", True)
        self.OIDC_NONCE_SIZE = self.get_settings("OIDC_NONCE_SIZE", 32)

    @staticmethod
    def get_settings(attr, *args):
        return import_from_settings(attr, *args)

    @cached_property
    def exempt_urls(self):
        """Generate and return a set of url paths to exempt from SessionRefresh

        This takes the value of ``settings.OIDC_EXEMPT_URLS`` and appends three
        urls that mozilla-django-oidc uses. These values can be view names or
        absolute url paths.

        :returns: list of url paths (for example "/oidc/callback/")

        """
        exempt_urls = []
        for url in self.OIDC_EXEMPT_URLS:
            if not isinstance(url, re_Pattern):
                exempt_urls.append(url)
        exempt_urls.extend(
            [
                "oidc_authentication_init",
                "oidc_authentication_callback",
                "oidc_logout",
            ]
        )

        return set(
            [url if url.startswith("/") else reverse(url) for url in exempt_urls]
        )

    @cached_property
    def exempt_url_patterns(self):
        """Generate and return a set of url patterns to exempt from SessionRefresh

        This takes the value of ``settings.OIDC_EXEMPT_URLS`` and returns the
        values that are compiled regular expression patterns.

        :returns: list of url patterns (for example,
            ``re.compile(r"/user/[0-9]+/image")``)
        """
        exempt_patterns = set()
        for url_pattern in self.OIDC_EXEMPT_URLS:
            if isinstance(url_pattern, re_Pattern):
                exempt_patterns.add(url_pattern)
        return exempt_patterns

    def is_refreshable_url(self, request):
        """Takes a request and returns whether it triggers a refresh examination

        :arg HttpRequest request:

        :returns: boolean

        """
        # Do not attempt to refresh the session if the OIDC backend is not used
        backend_session = request.session.get(BACKEND_SESSION_KEY)
        is_oidc_enabled = True
        if backend_session:
            auth_backend = import_string(backend_session)
            is_oidc_enabled = issubclass(auth_backend, OIDCAuthenticationBackend)

        return (
            request.method == "GET"
            and request.user.is_authenticated
            and is_oidc_enabled
            and request.path not in self.exempt_urls
            and not any(pat.match(request.path) for pat in self.exempt_url_patterns)
        )

    def process_request(self, request):
        if not self.is_refreshable_url(request):
            LOGGER.debug("request is not refreshable")
            return

        expiration = request.session.get("oidc_id_token_expiration", 0)
        now = time.time()
        if expiration > now:
            # The id_token is still valid, so we don't have to do anything.
            LOGGER.debug("id token is still valid (%s > %s)", expiration, now)
            return

        LOGGER.debug("id token has expired")
        # The id_token has expired, so we have to re-authenticate silently.
        auth_url = self.OIDC_OP_AUTHORIZATION_ENDPOINT
        client_id = self.OIDC_RP_CLIENT_ID
        state = get_random_string(self.OIDC_STATE_SIZE)

        # Build the parameters as if we were doing a real auth handoff, except
        # we also include prompt=none.
        params = {
            "response_type": "code",
            "client_id": client_id,
            "redirect_uri": absolutify(
                request, reverse(self.OIDC_AUTHENTICATION_CALLBACK_URL)
            ),
            "state": state,
            "scope": self.OIDC_RP_SCOPES,
            "prompt": "none",
        }

        params.update(self.get_settings("OIDC_AUTH_REQUEST_EXTRA_PARAMS", {}))

        if self.OIDC_USE_NONCE:
            nonce = get_random_string(self.OIDC_NONCE_SIZE)
            params.update({"nonce": nonce})

        if self.get_settings("OIDC_USE_PKCE", False):
            code_verifier_length = self.get_settings("OIDC_PKCE_CODE_VERIFIER_SIZE", 64)
            # Check that code_verifier_length is between the min and max length
            # defined in https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
            if not (43 <= code_verifier_length <= 128):
                raise ValueError("code_verifier_length must be between 43 and 128")

            # Generate code_verifier and code_challenge pair
            code_verifier = get_random_string(code_verifier_length)
            code_challenge_method = self.get_settings(
                "OIDC_PKCE_CODE_CHALLENGE_METHOD", "S256"
            )
            code_challenge = generate_code_challenge(
                code_verifier, code_challenge_method
            )

            # Append code_challenge to authentication request parameters
            params.update(
                {
                    "code_challenge": code_challenge,
                    "code_challenge_method": code_challenge_method,
                }
            )
        else:
            code_verifier = None

        add_state_and_verifier_and_nonce_to_session(
            request, state, params, code_verifier
        )

        request.session["oidc_login_next"] = request.get_full_path()

        query = urlencode(params, quote_via=quote)
        redirect_url = "{url}?{query}".format(url=auth_url, query=query)
        if request.headers.get("x-requested-with") == "XMLHttpRequest":
            # Almost all XHR request handling in client-side code struggles
            # with redirects since redirecting to a page where the user
            # is supposed to do something is extremely unlikely to work
            # in an XHR request. Make a special response for these kinds
            # of requests.
            # The use of 403 Forbidden is to match the fact that this
            # middleware doesn't really want the user in if they don't
            # refresh their session.
            response = JsonResponse({"refresh_url": redirect_url}, status=403)
            response["refresh_url"] = redirect_url
            return response
        return HttpResponseRedirect(redirect_url)



import logging
import time
import warnings
from hashlib import sha256
from urllib.request import parse_http_list, parse_keqv_list

# Make it obvious that these aren't the usual base64 functions
import josepy.b64
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

LOGGER = logging.getLogger(__name__)


def parse_www_authenticate_header(header):
    """
    Convert a WWW-Authentication header into a dict that can be used
    in a JSON response.
    """
    items = parse_http_list(header)
    return parse_keqv_list(items)


def import_from_settings(attr, *args):
    """
    Load an attribute from the django settings.

    :raises:
        ImproperlyConfigured
    """
    try:
        if args:
            return getattr(settings, attr, args[0])
        return getattr(settings, attr)
    except AttributeError:
        raise ImproperlyConfigured("Setting {0} not found".format(attr))


def absolutify(request, path):
    """Return the absolute URL of a path."""
    return request.build_absolute_uri(path)


def is_authenticated(user):
    """return True if the user is authenticated.
    This is necessary because in Django 1.10 the `user.is_authenticated`
    stopped being a method and is now a property.
    Actually `user.is_authenticated()` actually works, thanks to a backwards
    compat trick in Django. But in Django 2.0 it will cease to work
    as a callable method.
    """

    msg = "`is_authenticated()` is going to be removed in mozilla-django-oidc v 2.x"
    warnings.warn(msg, DeprecationWarning)
    return user.is_authenticated


def base64_url_encode(bytes_like_obj):
    """Return a URL-Safe, base64 encoded version of bytes_like_obj

    Implements base64urlencode as described in
    https://datatracker.ietf.org/doc/html/rfc7636#appendix-A
    """

    s = josepy.b64.b64encode(bytes_like_obj).decode("ascii")  # base64 encode
    # the josepy base64 encoder (strips '='s padding) automatically
    s = s.replace("+", "-")  # 62nd char of encoding
    s = s.replace("/", "_")  # 63rd char of encoding

    return s


def base64_url_decode(string_like_obj):
    """Return the bytes encoded in a URL-Safe, base64 encoded string.
    Implements inverse of base64urlencode as described in
    https://datatracker.ietf.org/doc/html/rfc7636#appendix-A
    This function is not used by the OpenID client; it's just for testing PKCE related functions.
    """
    s = string_like_obj

    s = s.replace("_", "/")  # 63rd char of encoding
    s = s.replace("-", "+")  # 62nd char of encoding
    b = josepy.b64.b64decode(s)  # josepy base64 encoder (decodes without '='s padding)

    return b


def generate_code_challenge(code_verifier, method):
    """Return a code_challege, which proves knowledge of the code_verifier.
    The code challenge is generated according to method which must be one
    of the methods defined in https://datatracker.ietf.org/doc/html/rfc7636#section-4.2:
    - plain:
    code_challenge = code_verifier
    - S256:
    code_challenge = BASE64URL-ENCODE(SHA256(ASCII(code_verifier)))
    """

    if method == "plain":
        return code_verifier

    elif method == "S256":
        return base64_url_encode(sha256(code_verifier.encode("ascii")).digest())

    else:
        raise ValueError("code challenge method must be 'plain' or 'S256'.")


def add_state_and_verifier_and_nonce_to_session(
    request, state, params, code_verifier=None
):
    """
    Stores the `state` and `nonce` parameters and an optional `code_verifier` (for PKCE) in a
    session dictionary which maps `state` -> {nonce, code_verifier}.  Each entry includes
    the time when it was added. The dictionary can contain multiple state -> {nonce, code_verifier}
    mappings to allow parallel logins with multiple browser sessions.
    To keep the session space to a reasonable size, the dictionary is kept at 50
    state -> {nonce, code_verifier} mappings maximum.
    """
    nonce = params.get("nonce")

    # OPs supporting PKCE will require `code_verifier` to be sent to the token
    # endpoint if `code_challenge` is sent to the authentication endpoint.
    # Make sure that `code_challenge` and `code_verifier` are both specified
    # or neither is.
    assert ("code_challenge" in params) == (code_verifier is not None)

    # Store Nonce with the State parameter in the session "oidc_states" dictionary.
    # The dictionary can store multiple State/Nonce combinations to allow parallel
    # authentication flows which would otherwise overwrite State/Nonce values!
    # The "oidc_states" dictionary uses the state as key and as value a dictionary with "nonce"
    # and "added_on". "added_on" contains the time when the state was added to the session.
    # With this value, the oldest element can be found and deleted from the session.
    if "oidc_states" not in request.session or not isinstance(
        request.session["oidc_states"], dict
    ):
        request.session["oidc_states"] = {}

    # Make sure that the State/Nonce dictionary in the session does not get too big.
    # If the number of State/Nonce combinations reaches a certain threshold, remove the oldest
    # state by finding out
    # which element has the oldest "add_on" time.
    limit = import_from_settings("OIDC_MAX_STATES", 50)
    if len(request.session["oidc_states"]) >= limit:
        LOGGER.info(
            'User has more than {} "oidc_states" in his session, '
            "deleting the oldest one!".format(limit)
        )
        oldest_state = None
        oldest_added_on = time.time()
        for item_state, item in request.session["oidc_states"].items():
            if item["added_on"] < oldest_added_on:
                oldest_state = item_state
                oldest_added_on = item["added_on"]
        if oldest_state:
            del request.session["oidc_states"][oldest_state]

    request.session["oidc_states"][state] = {
        "code_verifier": code_verifier,
        "nonce": nonce,
        "added_on": time.time(),
    }



import unittest.mock

from django.core.exceptions import SuspiciousOperation
from django.test import RequestFactory, TestCase, override_settings
from rest_framework import exceptions

from mozilla_django_oidc.contrib.drf import OIDCAuthentication


class TestDRF(TestCase):
    @override_settings(OIDC_OP_TOKEN_ENDPOINT="https://server.example.com/token")
    @override_settings(OIDC_OP_USER_ENDPOINT="https://server.example.com/user")
    @override_settings(OIDC_RP_CLIENT_ID="example_id")
    @override_settings(OIDC_RP_CLIENT_SECRET="client_secret")
    def setUp(self):
        self.auth = OIDCAuthentication(backend=unittest.mock.Mock())
        self.request = RequestFactory().get("/", HTTP_AUTHORIZATION="Bearer faketoken")

    def test_authenticate_returns_none_if_no_access_token(self):
        with unittest.mock.patch.object(self.auth, "get_access_token", return_value=None):
            ret = self.auth.authenticate(self.request)
        self.assertIsNone(ret)

    def test_authenticate_raises_authenticationfailed_if_backend_returns_no_user(self):
        self.auth.backend.get_or_create_user.return_value = None
        with self.assertRaises(exceptions.AuthenticationFailed):
            self.auth.authenticate(self.request)

    def test_authenticate_raises_authenticationfailed_on_suspiciousoperation(self):
        self.auth.backend.get_or_create_user.side_effect = SuspiciousOperation
        with self.assertRaises(exceptions.AuthenticationFailed):
            self.auth.authenticate(self.request)

    def test_returns_user_and_token_if_backend_returns_user(self):
        user = unittest.mock.Mock()
        self.auth.backend.get_or_create_user.return_value = user
        ret = self.auth.authenticate(self.request)
        self.assertEqual(ret[0], user)
        self.assertEqual(ret[1], "faketoken")
